import type { DataTag, QueryClient, QueryFilters } from '@tanstack/react-query';
import type { TRPCClient, TRPCClientErrorLike } from '@trpc/client';
import { TRPCUntypedClient } from '@trpc/client';
import type { AnyTRPCProcedure, AnyTRPCRootTypes, AnyTRPCRouter, inferProcedureInput, inferRouterContext, inferTransformedProcedureOutput, TRPCProcedureType, TRPCRouterRecord } from '@trpc/server';
import type { MaybePromise } from '@trpc/server/unstable-core-do-not-import';
import { type TRPCInfiniteQueryOptions } from './infiniteQueryOptions';
import type { MutationOptionsOverride } from './mutationOptions';
import { type TRPCMutationOptions } from './mutationOptions';
import { type TRPCQueryOptions } from './queryOptions';
import { type TRPCSubscriptionOptions } from './subscriptionOptions';
import type { OptionalCursorInput, ResolverDef, TRPCInfiniteData, TRPCMutationKey, TRPCQueryKey, WithRequired } from './types';
export interface DecorateRouterKeyable {
    /**
     * Calculate the TanStack Query Key for any path, could be used to invalidate every procedure beneath this path
     *
     * @see https://tanstack.com/query/latest/docs/framework/react/guides/query-keys
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#queryKey
     */
    pathKey: () => TRPCQueryKey;
    /**
     * Calculate a TanStack Query Filter for any path, could be used to manipulate every procedure beneath this path
     *
     * @see https://tanstack.com/query/latest/docs/framework/react/guides/filters
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#queryFilter
     */
    pathFilter: (filters?: QueryFilters<TRPCQueryKey>) => WithRequired<QueryFilters<TRPCQueryKey>, 'queryKey'>;
}
interface TypeHelper<TDef extends ResolverDef> {
    /**
     * @internal prefer using inferInput and inferOutput to access types
     */
    '~types': {
        input: TDef['input'];
        output: TDef['output'];
        errorShape: TDef['errorShape'];
    };
}
export type inferInput<TProcedure extends DecorateInfiniteQueryProcedure<any> | DecorateQueryProcedure<any> | DecorateMutationProcedure<any>> = TProcedure['~types']['input'];
export type inferOutput<TProcedure extends DecorateInfiniteQueryProcedure<any> | DecorateQueryProcedure<any> | DecorateMutationProcedure<any>> = TProcedure['~types']['output'];
export interface DecorateInfiniteQueryProcedure<TDef extends ResolverDef> extends TypeHelper<TDef> {
    /**
     * Create a set of type-safe infinite query options that can be passed to `useInfiniteQuery`, `prefetchInfiniteQuery` etc.
     *
     * @see https://tanstack.com/query/latest/docs/framework/react/reference/infiniteQueryOptions#infinitequeryoptions
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#infiniteQueryOptions
     */
    infiniteQueryOptions: TRPCInfiniteQueryOptions<TDef>;
    /**
     * Calculate the TanStack Query Key for a Infinite Query Procedure
     *
     * @see https://tanstack.com/query/latest/docs/framework/react/guides/query-keys
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#queryKey
     */
    infiniteQueryKey: (input?: Partial<TDef['input']>) => DataTag<TRPCQueryKey, TRPCInfiniteData<TDef['input'], TDef['output']>, TRPCClientErrorLike<{
        transformer: TDef['transformer'];
        errorShape: TDef['errorShape'];
    }>>;
    /**
     * Calculate a TanStack Query Filter for a Infinite Query Procedure
     *
     * @see https://tanstack.com/query/latest/docs/framework/react/guides/filters
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#queryFilter
     */
    infiniteQueryFilter: (input?: Partial<TDef['input']>, filters?: QueryFilters<DataTag<TRPCQueryKey, TRPCInfiniteData<TDef['input'], TDef['output']>, TRPCClientErrorLike<{
        transformer: TDef['transformer'];
        errorShape: TDef['errorShape'];
    }>>>) => WithRequired<QueryFilters<DataTag<TRPCQueryKey, TRPCInfiniteData<TDef['input'], TDef['output']>, TRPCClientErrorLike<{
        transformer: TDef['transformer'];
        errorShape: TDef['errorShape'];
    }>>>, 'queryKey'>;
}
export interface DecorateQueryProcedure<TDef extends ResolverDef> extends TypeHelper<TDef>, DecorateRouterKeyable {
    /**
     * Create a set of type-safe query options that can be passed to `useQuery`, `prefetchQuery` etc.
     *
     * @see https://tanstack.com/query/latest/docs/framework/react/reference/queryOptions#queryoptions
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#queryOptions
     */
    queryOptions: TRPCQueryOptions<TDef>;
    /**
     * Calculate the TanStack Query Key for a Query Procedure
     *
     * @see https://tanstack.com/query/latest/docs/framework/react/guides/query-keys
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#queryKey
     */
    queryKey: (input?: Partial<TDef['input']>) => DataTag<TRPCQueryKey, TDef['output'], TRPCClientErrorLike<{
        transformer: TDef['transformer'];
        errorShape: TDef['errorShape'];
    }>>;
    /**
     * Calculate a TanStack Query Filter for a Query Procedure
     *
     * @see https://tanstack.com/query/latest/docs/framework/react/guides/filters
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#queryFilter
     */
    queryFilter: (input?: Partial<TDef['input']>, filters?: QueryFilters<DataTag<TRPCQueryKey, TDef['output'], TRPCClientErrorLike<{
        transformer: TDef['transformer'];
        errorShape: TDef['errorShape'];
    }>>>) => WithRequired<QueryFilters<DataTag<TRPCQueryKey, TDef['output'], TRPCClientErrorLike<{
        transformer: TDef['transformer'];
        errorShape: TDef['errorShape'];
    }>>>, 'queryKey'>;
}
export interface DecorateMutationProcedure<TDef extends ResolverDef> extends TypeHelper<TDef> {
    /**
     * Create a set of type-safe mutation options that can be passed to `useMutation`
     *
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#mutationOptions
     */
    mutationOptions: TRPCMutationOptions<TDef>;
    /**
     * Calculate the TanStack Mutation Key for a Mutation Procedure
     *
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#mutationKey
     */
    mutationKey: () => TRPCMutationKey;
}
export interface DecorateSubscriptionProcedure<TDef extends ResolverDef> {
    /**
     * Create a set of type-safe subscription options that can be passed to `useSubscription`
     *
     * @see https://trpc.io/docs/client/tanstack-react-query/usage#subscriptionOptions
     */
    subscriptionOptions: TRPCSubscriptionOptions<TDef>;
}
export type DecorateProcedure<TType extends TRPCProcedureType, TDef extends ResolverDef> = TType extends 'query' ? DecorateQueryProcedure<TDef> & (TDef['input'] extends OptionalCursorInput ? DecorateInfiniteQueryProcedure<TDef> : Record<string, never>) : TType extends 'mutation' ? DecorateMutationProcedure<TDef> : TType extends 'subscription' ? DecorateSubscriptionProcedure<TDef> : never;
/**
 * @internal
 */
export type DecoratedRouterRecord<TRoot extends AnyTRPCRootTypes, TRecord extends TRPCRouterRecord> = {
    [TKey in keyof TRecord]: TRecord[TKey] extends infer $Value ? $Value extends TRPCRouterRecord ? DecoratedRouterRecord<TRoot, $Value> & DecorateRouterKeyable : $Value extends AnyTRPCProcedure ? DecorateProcedure<$Value['_def']['type'], {
        input: inferProcedureInput<$Value>;
        output: inferTransformedProcedureOutput<TRoot, $Value>;
        transformer: TRoot['transformer'];
        errorShape: TRoot['errorShape'];
    }> : never : never;
};
export type TRPCOptionsProxy<TRouter extends AnyTRPCRouter> = DecoratedRouterRecord<TRouter['_def']['_config']['$types'], TRouter['_def']['record']> & DecorateRouterKeyable;
export interface TRPCOptionsProxyOptionsBase {
    queryClient: QueryClient | (() => QueryClient);
    overrides?: {
        mutations?: MutationOptionsOverride;
    };
}
export interface TRPCOptionsProxyOptionsInternal<TRouter extends AnyTRPCRouter> {
    router: TRouter;
    ctx: inferRouterContext<TRouter> | (() => MaybePromise<inferRouterContext<TRouter>>);
}
export interface TRPCOptionsProxyOptionsExternal<TRouter extends AnyTRPCRouter> {
    client: TRPCUntypedClient<TRouter> | TRPCClient<TRouter>;
}
export type TRPCOptionsProxyOptions<TRouter extends AnyTRPCRouter> = TRPCOptionsProxyOptionsBase & (TRPCOptionsProxyOptionsInternal<TRouter> | TRPCOptionsProxyOptionsExternal<TRouter>);
/**
 * Create a typed proxy from your router types. Can also be used on the server.
 *
 * @see https://trpc.io/docs/client/tanstack-react-query/setup#3b-setup-without-react-context
 * @see https://trpc.io/docs/client/tanstack-react-query/server-components#5-create-a-trpc-caller-for-server-components
 */
export declare function createTRPCOptionsProxy<TRouter extends AnyTRPCRouter>(opts: TRPCOptionsProxyOptions<TRouter>): TRPCOptionsProxy<TRouter>;
export {};
//# sourceMappingURL=createOptionsProxy.d.ts.map